import json
import os
from compression import zstd
from pathlib import Path
from typing import Any, Callable

from ocsf_schema_compiler.exceptions import SchemaException
from ocsf_schema_compiler.jsonish import (
    JObject,
    json_type_from_value,
    j_object,
    j_string,
)


# _load_json_object_file loads a JSON file an ensure the result is a JSON object.
# It also has all of the ugly Pyright annotations to deal with the loose typing of
# open() and json.load().
def _load_json_object_file(
    path: Path,
    f: Any,  # pyright: ignore[reportAny, reportExplicitAny]
) -> JObject:
    v: Any = json.load(f)  # pyright: ignore[reportAny, reportExplicitAny]
    if not isinstance(v, dict):
        t = json_type_from_value(v)  # pyright: ignore[reportAny]
        raise TypeError(
            f"Schema file contains a JSON {t} value, but should contain an object:"
            f" {path}"
        )
    return v  # pyright: ignore[reportUnknownVariableType]


def read_json_object_file(path: Path) -> JObject:
    with open(path) as f:
        return _load_json_object_file(path, f)


def read_json_object_zstandard_file(path: Path) -> JObject:
    with zstd.open(path) as f:
        return _load_json_object_file(path, f)


def read_structured_items(
    base_path: Path,
    kind: str,
    item_callback_fn: Callable[[Path, JObject], None] | None = None,
) -> JObject:
    """
    Read schema structured items found in `kind` directory under `base_path`,
    recursively, and returns dict with unprocessed items, each keyed by their name
    attribute.
    """
    # event classes can be organized in subdirectories, so we must walk to find all the
    # event class JSON files
    item_path = base_path / kind
    items: JObject = {}
    for dir_path, _dir_names, file_names in os.walk(item_path, topdown=False):
        for file_name in file_names:
            if file_name.endswith(".json"):
                file_path = Path(dir_path, file_name)
                obj = read_json_object_file(file_path)
                name = obj.get("name")

                # The way this is tested, "no value" happens when attribute is missing,
                # JSON null (Python None), or an empty value (an empty string, JSON
                # array, JSON object, or even a numeric zero).
                if not name:
                    raise SchemaException(
                        f'The "name" value in {kind} file must have a value:'
                        f" {file_path}"
                    )

                # Ensure name is a string
                if not isinstance(name, str):
                    raise SchemaException(
                        f'The "name" value in {kind} file must be a string,'
                        f" but got {json_type_from_value(name)}: {file_path}"
                    )

                if name in items:
                    existing = j_object(items[name])
                    raise SchemaException(
                        f'Collision of "name" in {kind} file: "{name}" with caption'
                        f' "{obj.get("caption", "")}", collides with {kind} with'
                        f' caption "{existing.get("caption", "")}", file: {file_path}'
                    )
                else:
                    items[name] = obj
                    if item_callback_fn:
                        item_callback_fn(file_path, obj)

    return items


def read_patchable_structured_items(
    base_path: Path,
    kind: str,
    item_callback_fn: Callable[[Path, JObject], None] | None = None,
) -> tuple[JObject, JObject]:
    """
    Read schema "patchable" structured items found in `kind` directory under
    `base_path`, recursively, and returns dataclass with unprocessed items and patches.
    Extension classes and objects are patchable structured items. Items are each keyed
    by their name attribute and patches are keyed by the name of the item to patch.

    Returns tuple of items dictionary and patches dictionary.
    """
    # event classes can be organized in subdirectories, so we must walk to find all the
    # event class JSON files
    item_path = base_path / kind
    items: JObject = {}
    patches: JObject = {}
    for dir_path, _dir_names, file_names in os.walk(item_path, topdown=False):
        for file_name in file_names:
            if file_name.endswith(".json"):
                file_path = Path(dir_path, file_name)
                obj = read_json_object_file(file_path)
                # An extension "patch" occurs in two cases:
                #   1. The item has an "extends" key but no "name" key. This is the
                #      common case in practice.
                #   2. The item has both the "name" and "extends" keys, and both have
                #      the same value.
                name = obj.get("name")
                extends = obj.get("extends")

                # A structured item (a class, object, etc.) must have a name OR an
                # extends value. The way this is tested, "no value" happens when
                # attribute is missing, JSON null (Python None), or an empty value (an
                # empty string, JSON array, JSON object, or even a numeric zero).
                if not name and not extends:
                    raise SchemaException(
                        f'Extension {kind} file does not have a "name" or "extends"'
                        f" value: {file_path}"
                    )

                # Ensure values are strings
                if name is not None and not isinstance(name, str):
                    raise SchemaException(
                        f'The "name" value in extension {kind} file must be a string,'
                        f" but got {json_type_from_value(name)}: {file_path}"
                    )
                if extends is not None and not isinstance(extends, str):
                    raise SchemaException(
                        f'The "extends" value in extension {kind} file must be a'
                        f" string, but got {json_type_from_value(extends)}: {file_path}"
                    )

                if not name or name == extends:
                    # This is a patch definition.
                    # An extension event class or object is a patch when it only defines
                    # "extends" or when "name" and "extends" have the same value. This
                    patch_name = j_string(extends)  # use patch_name for clarity
                    if patch_name in patches:
                        existing = j_object(patches[patch_name])
                        raise SchemaException(
                            f'Collision of patch name ("extends" key) in extension'
                            f' {kind} file: "{patch_name}" with caption'
                            f' "{obj.get("caption", "")}", collides with existing'
                            f' {kind} with caption "{existing.get("caption", "")}",'
                            f" file: {file_path}"
                        )
                    else:
                        patches[patch_name] = obj
                        if item_callback_fn:
                            item_callback_fn(file_path, obj)
                else:
                    # This is a normal definition.
                    if name in items:
                        existing = j_object(items[name])
                        raise SchemaException(
                            f'Collision of "name" in extension {kind} file: "{name}"'
                            f' with caption "{obj.get("caption", "")}", collides with'
                            f' {kind} with caption "{existing.get("caption", "")}",'
                            f" file: {file_path}"
                        )
                    else:
                        items[name] = obj
                        if item_callback_fn:
                            item_callback_fn(file_path, obj)

    return items, patches



import json
from copy import deepcopy
from dataclasses import dataclass
from typing import Callable

from ocsf_schema_compiler.jsonish import JObject, JValue


class Missing:
    pass


MISSING = Missing()


@dataclass
class DiffDictKeys:
    keys: list[str]
    other_name: str


type DiffValue = Missing | DiffDictKeys | JValue


@dataclass
class Difference:
    is_expected: bool
    path: list[str]
    left: DiffValue
    right: DiffValue

    def formatted_string(self) -> str:
        if self.is_expected:
            kind = "expected"
        else:
            kind = "unexpected"
        return (
            f'Diff at "{_path_to_string(self.path)}" ({kind}):'
            f"\n    left  : {_diff_value_to_string(self.left)}"
            f"\n    right : {_diff_value_to_string(self.right)}"
        )


def formatted_diffs(diffs: list[Difference]) -> str:
    # Show unexpected before expected
    unexpected_diffs: list[str] = []
    expected_diffs: list[str] = []
    for diff in diffs:
        if diff.is_expected:
            expected_diffs.append(diff.formatted_string())
        else:
            unexpected_diffs.append(diff.formatted_string())
    return "\n".join(unexpected_diffs + expected_diffs)


type DiffCallback = Callable[
    [str, list[str], JObject, JObject, DiffValue, DiffValue], bool
]


def diff_objects(
    left: JObject, right: JObject, diff_callback: DiffCallback | None = None
) -> tuple[bool, list[Difference]]:
    diffs: list[Difference] = []
    _diff_objects(left, right, diff_callback, [], diffs)
    ok = True
    for diff in diffs:
        if not diff.is_expected:
            ok = False
            break
    return ok, diffs


def _diff_objects(
    left_obj: JObject,
    right_obj: JObject,
    diff_callback: DiffCallback | None,
    base_path: list[str],
    diffs: list[Difference],
) -> None:
    for key in sorted(set(left_obj.keys()) | set(right_obj.keys())):
        path = deepcopy(base_path)
        path.append(key)
        left_dv = _diff_get(left_obj, key)
        right_dv = _diff_get(right_obj, key)
        if isinstance(left_dv, dict) and isinstance(right_dv, dict):
            has_equal_key, left_key_dv, right_key_dv = _is_equal_keys(left_dv, right_dv)
            if has_equal_key:
                _diff_objects(left_dv, right_dv, diff_callback, path, diffs)
            else:
                _callback_append(
                    key,
                    path,
                    left_obj,
                    right_obj,
                    left_key_dv,
                    right_key_dv,
                    diff_callback,
                    diffs,
                )
        elif left_dv != right_dv:
            _callback_append(
                key,
                path,
                left_obj,
                right_obj,
                left_dv,
                right_dv,
                diff_callback,
                diffs,
            )


def _callback_append(
    key: str,
    path: list[str],
    left_obj: JObject,
    right_obj: JObject,
    left_dv: DiffValue,
    right_dv: DiffValue,
    diff_callback: DiffCallback | None,
    diffs: list[Difference],
) -> None:
    if diff_callback:
        is_expected = diff_callback(key, path, left_obj, right_obj, left_dv, right_dv)
        diffs.append(Difference(is_expected, path, left_dv, right_dv))
    else:
        diffs.append(Difference(False, path, left_dv, right_dv))


def _is_equal_keys(
    left_obj: JObject, right_obj: JObject
) -> tuple[bool, DiffValue, DiffValue]:
    left_keys = set(left_obj.keys())
    right_keys = set(right_obj.keys())
    if left_keys == right_keys:
        return True, None, None
    else:
        return (
            False,
            DiffDictKeys(sorted(left_keys - right_keys), other_name="right"),
            DiffDictKeys(sorted(right_keys - left_keys), other_name="left"),
        )


def _diff_get(o: JObject, k: str) -> DiffValue:
    if k in o:
        return o[k]
    return MISSING


def _path_to_string(path: list[str]) -> str:
    return ".".join(path)


def _diff_value_to_string(dv: DiffValue) -> str:
    if isinstance(dv, Missing):
        return "missing"
    if isinstance(dv, DiffDictKeys):
        if dv.keys:
            return f"key(s) not in {dv.other_name} object: {', '.join(dv.keys)}"
        else:
            return f"key(s) not in {dv.other_name} object: none"
    return json.dumps(dv, sort_keys=True)



# Type aliases for JSON-compatible types. See https://json.org.
# Yes, these are circular, and Python is OK with that.
# As with all Python type hints, these improve code readability and help IDEs identify
# type mismatches.

# JValue is type alias for types compatible with JSON values.
type JValue = JObject | JArray | str | int | float | bool | None
# JObject is a type alias for dictionary compatible with a JSON object.
type JObject = dict[str, JValue]
# JArray is a type alias for types compatible with a JSON array.
type JArray = list[JValue]  # TODO | tuple[JValue]


def json_type_from_value(value: object) -> str:
    """
    Return JSON type for a Python value. See https://json.org.
    This is intended for error messages.
    """
    if isinstance(value, dict):
        return "object"
    if isinstance(value, list):
        return "array"
    if isinstance(value, str):
        return "string"
    if isinstance(value, int):
        return "number (integer)"
    if isinstance(value, float):
        return "number (float)"
    if isinstance(value, bool):
        if value:
            return "true"
        return "false"
    if value is None:
        return "null"
    return f"non-JSON type: {type(value).__name__}"


# These j_* function are for type safety. They keep Pyright happy.
# The assertion error messages are given in terms of JSON types, mostly.
# JSON does not have a integer type, though we have j_integer rather than a more
# general j_number.


def j_object(v: JValue) -> JObject:
    """Ensures value v is a JObject (dict) and returns it."""
    assert isinstance(v, dict), (
        f"j_object: expected object but got {json_type_from_value(v)}: {v}"
    )
    return v


def j_object_optional(v: JValue) -> JObject | None:
    """Ensures value v is a JObject (dict) or None and returns it."""
    assert v is None or isinstance(v, dict), (
        f"j_object: expected object or null but got {json_type_from_value(v)}: {v}"
    )
    return v


def j_array(v: JValue) -> JArray:
    """Ensures value v is a JArray (list) and returns it."""
    assert isinstance(v, list), (
        f"j_array: expected array but got {json_type_from_value(v)}: {v}"
    )
    return v


def j_array_optional(v: JValue) -> JArray | None:
    """Ensures value v is a JArray (list) or None and returns it."""
    assert v is None or isinstance(v, list), (
        f"j_array: expected array or null but got {json_type_from_value(v)}: {v}"
    )
    return v


def j_string(v: JValue) -> str:
    """Ensures value v is a string and returns it."""
    assert isinstance(v, str), (
        f"j_string: expected string but got {json_type_from_value(v)}: {v}"
    )
    return v


def j_string_optional(v: JValue) -> str | None:
    """Ensures value v is a string or None and returns it."""
    assert v is None or isinstance(v, str), (
        f"j_string: expected string or null but got {json_type_from_value(v)}: {v}"
    )
    return v


def j_integer(v: JValue) -> int:
    """Ensures value v is an integer and returns it."""
    assert isinstance(v, int), (
        f"j_integer: expected integer number but got {json_type_from_value(v)}: {v}"
    )
    return v
